#include <iostream>
#include <sstream>
#include <vector>
#include <cstdio>
#include <cstdlib>
#include "midi.hpp"
#include "midid.hpp"
#include "mmlgen.h"
#include "option.h"

namespace MIDI {
static const char *note_table[] = {
	"c", "c+", "d", "d+", "e", "f",
	"f+", "g", "g+", "a", "a+", "b",
	"r",
};

static const char *track_string[] = {
	"#A","#B","#C","#D","#E","#F","#G","#H","#I","#J",
	"#K","#L","#M","#N","#O","#P","#Q","#R","#S","#T",
	"#U","#V","#W","#X","#Y","#Z",
};

static const struct
{
	int tick;
	const char *len;
} len_table[] = {
	{ 1920, "1",   },
	{ 1680, "2..", },
	{ 1440, "2.",  },
	{ 960 , "2",   },
	{ 840 , "4..", },
	{ 720 , "4.",  },
	{ 480 , "4",   },
	{ 420 , "8..", },
	{ 360 , "8.",  },
	{ 240 , "8",   },
	{ 210 , "16..",},
	{ 180 , "16.", },
	{ 120 , "16",  },
	{ 90  , "32.", },
	{ 60  , "32",  },
	{ 30  , "64",  },
	{ 15  , "128", },
	{ 10  , "192", },
	{ 8   , "240", },
	{ 5   , "384", },
	{ 4   , "480", },
	{ 2   , "960", },
	{ 1   ,"1920", },
};         

static const char* mml_rest         = "r";
static const char* mml_length       = "l";
static const char* mml_tie          = "^";
static const char* mml_octave       = "o";
static const char* mml_quant        = "q";
static const char* mml_at_quant     = "@q";
static const char* mml_volume       = "v";
static const char* mml_at_volume    = "@v";
static const char* mml_detune       = "k";
static const char* mml_panpot       = "v";
static const char* mml_at_panpot    = "@p";
static const char* mml_mod_type     = "%";
static const char* mml_mod_param    = "@";
static const char* mml_input        = "@i";
static const char* mml_output       = "@o";
static const char* mml_volumeUp     = "(";
static const char* mml_volumeDown   = ")";
static const char* mml_octaveUp     = "<";
static const char* mml_octaveDown   = ">";
static const char* mml_slur         = "&";
static const char* mml_slurweak     = "&&";
static const char* mml_portament    = "*";
static const char* mml_parameter    = ",";
static const char* mml_end_sequence = ";";
static const char* mml_repeatPoint  = "$";
static const char* mml_repeatBegin  = "[";
static const char* mml_repeatEnd    = "]";
static const char* mml_repeatBreak  = "|";
static const char* mml_noteShift    = "!@ns";
static const char* mml_tempo        = "t";
static const char* mml_endmarker    = ";";

// int max_length = 1920;

static void print_length(std::ostringstream& str, int clock, size_t length, int tie_flag)
{
	while (clock >= len_table[length].tick) 
	{
		if (tie_flag) 
		{
			str << mml_tie;
		}
		str << len_table[length].len;
		clock -= len_table[length].tick;
		tie_flag = 1;
	}
	if (clock > 0) 
	{
		if (length >= sizeof(len_table) / sizeof(len_table[0])) 
		{
			return;
		} 
		else 
		{
			print_length(str, clock, length + 1, tie_flag);
		}
	}
	return;
}

bool 
mmlgen(std::ostringstream& str, DMIDI* dm, Status& sta)
{
/*
	unsigned int last_clock = 0;
	unsigned int last_note_clock = 0;
	unsigned int last_note = 0xff;
	unsigned int vel = 0xff;
	unsigned int last_vel = 0xff;
	int oct_base = 0xff;
	bool note_on = false;
	bool note_printed = false;
	bool tied = false;
	bool tie_remain = false;
	bool module_type_out = false;
*/
	const char *od, *ou;
	if (OPTION::enable_octave_rev) 
	{
		od = ">";
		ou = "<";
	}
	else 
	{
		od = "<";
		ou = ">";
	}

	std::vector<DMIDI_EVENT>::iterator it;
	if (sta.beg == std::vector<DMIDI_EVENT>::iterator(0))
	{
		str << track_string[sta.track] << "= ";
		it = dm->events.begin();
	}
	else
	{
		str << track_string[sta.track] << "+= ";
		it = sta.beg;
	}
	std::vector<DMIDI_EVENT>::iterator end = dm->events.end();
	while (it != end)
	{
		const unsigned int clock = it->clock;
//		const unsigned int ch = it->ch;
		const unsigned int mes = it->mes;
		const unsigned int lsb = it->lsb;
		const unsigned int msb = it->msb;

		// NbNił特i߂
		if (clock > sta.last_clock) 
		{
			if (sta.note_on)
			{
				// xVeB
				if (sta.last_vel != sta.vel && OPTION::enable_vel)
				{
					str << mml_volume << (sta.vel / 8);
					sta.last_vel = sta.vel;
				}

				int oct = (sta.last_note / 12);
				int oct_ = oct;
				if (oct != sta.oct_base) 
				{
					if (std::abs(oct - sta.oct_base) >= 2) 
					{
						str << "o" << oct;
					} else 
					if (oct < sta.oct_base) 
					{
						while(oct_ < sta.oct_base) 
						{
							str << od;
							oct_++;
						}
					}
					else
					{
						while(oct_ > sta.oct_base) 
						{
							str << ou;
							oct_--;
						}
					}
					sta.oct_base = oct;
				}
				sta.note_printed = true;
			}
			sta.tied = false;

			// ߂܂ꍇ͎̏߂܂Ői񂾂̂s
			if ((clock / 1920) > (sta.last_clock / 1920))
			{
				if (!sta.note_on)
					str << mml_rest;
				else
					str << note_table[sta.last_note % 12];
				int remain = 1920 - sta.last_clock % 1920;
				sta.last_clock += remain;
				print_length(str, remain, 0, 0);
				if (clock > sta.last_clock && sta.note_on) 
					str << mml_slur;
				str << ";" << std::endl;
				sta.beg = it;
				return false;
			}
/*
			while (clock - sta.last_clock > 1920)
			{
				if (!sta.note_on)
					str << mml_rest;
				else
					str << note_table[sta.last_note % 12];
				print_length(str, 1920, 0, 0);
				sta.last_clock += 1920;
				if (clock > sta.last_clock && sta.note_on) 
					str << mml_slur;
				str << std::endl;
			}
*/
			if (clock > sta.last_clock)
			{
				if (!sta.note_on)
					str << mml_rest;
				else
					str << note_table[sta.last_note % 12];
				print_length(str, clock - sta.last_clock, 0, 0);
			}
		}

		// m[gIƓ CC ȂǂĂꍇ
		// note_printed ^ɂȂ܂Ń^Cۗ
		if (sta.note_on 
			&& !sta.tied 
			&& (mes & 0xf0) != NOTE_ON 
			&& (mes & 0xf0) != NOTE_OFF
			&& sta.note_printed)
		{
			str << mml_slur;
			sta.tied = true;
		}

		switch (mes & 0xf0) 
		{
			case DMIDI_SYS: 
			{
				switch (mes) 
				{
					// e|
					case DMIDI_SYS_BPM: 
					{
						unsigned int bpm = lsb + (msb << 8);
						str << mml_tempo << bpm;
						break;
					}

					// I[}[J[
					case DMIDI_SYS_ENDDATA: 
					{
						str << mml_endmarker << std::endl;
						break;
					}
				}
				break;
			}

			case PITCHBEND:
			{
				int pitch = static_cast<short>(lsb | msb << 8);
				str << mml_detune << pitch;
				break;
			}

			case PC: 
			{
				if (!sta.module_type_out)
				{
					str << "q8" << mml_mod_type << 6;
					sta.module_type_out = true;
				}
				str << mml_mod_param << lsb;

				sta.oct_base = 0xff; // ProgramChange͊mɃIN^[uo͂
				break;
			}

			case CC: 
			{
				if (lsb == LOOP_POINT)
				{
					str << " $ ";
				}
				if (lsb == VOLUME)
				{
					// GNXvbVVOLUMEɓĂ邽
					// {[Cɂ΂悢
					str << mml_at_volume << msb;
				}
				if (lsb == PANPOT)
				{
					str << mml_at_panpot << (msb - 64);
				}
				break;
			}

			case NOTE_ON: 
			{
/*				if (last_note != lsb && note_on && !tied)
				{
					str << mml_slur;
				}
*/
				sta.tied = false;
				sta.tie_remain = false;
				sta.note_printed = false;

				sta.vel = msb;
				sta.note_on = true;
				sta.last_note = lsb;
				sta.last_note_clock = clock;
				break;
			}

			case NOTE_OFF: 
			{
				sta.note_on = false;
				sta.last_note = 0xff;
				sta.note_printed = false;
				break;
			}
			
			default: 
			{
				break;
			}
		}
		sta.last_clock = clock;
		++it;
	}
	return true;
}

} // namespace MID
